//  RW Foundations
// ================
// The RW Foundational Libraries

// Filename: ./include/incls_all.hpp
// -The following lines pull in the project's real *.hpp files.
// -note: this file also acts as a location for a few utility features

#pragma once

#include <algorithm>
#include <chrono>
#include <cstdio>
#include <iomanip>
#include <iostream>
#include <memory>
#include <ratio>
#include <sstream>
#include <stack>
#include <stdexcept>
#include <string>
#include <vector>

using std::any_of;
using std::cbegin;
using std::cend;
using std::cerr;
using std::cout;
using std::dec;
using std::exception;
using std::fixed;
using std::hex;
using std::localtime;
using std::make_unique;
using std::ostringstream;
using std::printf;
using std::put_time;
using std::puts;
using std::ratio;
using std::ratio_multiply;
using std::runtime_error;
using std::setprecision;
using std::setw;
using std::stack;
using std::string;
using std::to_string;
using std::unique_ptr;
using std::vector;
using std::chrono::duration;
using std::chrono::duration_cast;
using std::chrono::hours;
using std::chrono::microseconds;
using std::chrono::milliseconds;
using std::chrono::minutes;
using std::chrono::nanoseconds;
using std::chrono::seconds;
using std::chrono::system_clock;
using std::chrono::time_point;

///

// constexpr bool use_stamps{false};
constexpr bool use_stamps{true};

//
// NOTE: It is probably a good idea to be cautious about moving things around
// hereafter within this file; some dependencies here are sensitive to ordering.
//

//------------------------------------------------------------------------------
/** Our alias for a @c void*
 */
using Address = void*;

//------------------------------------------------------------------------------
/** Treat an @c Address as a reference to a general type @a T
 * @param[in] pt the type's @c Address to take the reference to
 * @return @a pt 's reference
 */
template <class T>
T& reference_to(Address pt) {
  return *static_cast<T*>(pt);
}

//------------------------------------------------------------------------------
/** Our alias for the value returned by various @c string member functions if
 * they fail.
 */
inline auto str_end = std::string::npos;

//------------------------------------------------------------------------------
#include "rw/rw_sb.hpp"  // RW class' fwd decl/alias

//------------------------------------------------------------------------------
/** The collection of pointers for all @c RW objects in the system
 * @note: stored-to during the object's constructor
 */
inline vector<RW*> RW_objs;

//------------------------------------------------------------------------------
namespace rw {
/** The type of a RW Foundations class
 */
enum class Type {
  Action,
  Analogy,
  Animeyes,
  Arigato,
  Artstory,
  Atashi,
  Bones,
  Bumpmaster,
  Chibitsu,
  Cognate,
  Concept,
  Curator,
  Dollhouse,
  Electro,
  Face,
  Gears,
  Graphicom,
  Hearsay,
  Meido,
  Moveit,
  Percept,
  Prop,
  Scene,
  Shell,
  Terminal,
  Waifusearch,
  //
  Foo  // sentinel for a general/unspecified type
};
}  // namespace rw

using rw::Type;

//------------------------------------------------------------------------------
/** A @c RW object's unique 'tag'. It is intended as a read-only identification
 * structure for all instantiated @c RW objects. It has these data elements:
 *
 * @b Handle: *human-readable string; it's 'name-tag'*
 *
 * @b ID: *doled-number, one for each new @c RW object created by the system*
 *
 * @b Pointer: *@c RW base object's (initial) location in the computer's memory*
 *
 * @b Type: *enumeration's item specification from the most important rw::Types*
 *
 * This is a tagging mechanism intended to be suitable for understanding the
 * (potentially highly-complex) callgraphs devisable for *RW Foundations*
 * systems. IE, @a 'who-called-whom-and-who-answered-the-call?'.
 *
 * It also forms the basic key for function-level access-control; to help ensure
 * against unintentional code misuse. For example, to allow for member objects
 * to only receive instructions from their owning object, since that owner
 * already knows the proper way to utilize that member's features correctly.
 * @note: The type name is an acronym: @b H andle, @b I d, @b P ointer,  @a (+)
 */
struct Hip {
  std::string   hdl{"nobody"};   // sentinel for an unassigned
  unsigned long id{9'001'42};    //  "
  RW*           ptr{nullptr};    //  "
  Type          typ{Type::Foo};  //  "

  /** Checks if this @c Hip tag's handle contains the supplied @a pattern
   * @param[in] pattern sub-string to locate
   * @return true if @c Hip::hdl member field contains @a pattern
   */
  inline auto contains(std::string const& pattern) const -> bool {
    // TODO: preload a 'to_lower()' on RW 's hip_.hdl in it's member init list
    // - also do a to_lower() on 'pattern' here
    return (hdl.find(pattern) != str_end);
  };

  /** Checks if this @c Hip tag's type is a match with the supplied @a type
   * @param[in] type the @c Type to check against
   * @return true if @c Hip::typ member field is equivalent to @a type
   */
  inline auto type_of(Type const type) const -> bool { return typ == type; }

  /** Deep check for an exact equivalency between this, and another @c Hip
   * @param other
   * @return true if an exact match
   */
  inline auto operator==(Hip const& other) const -> bool {
    return (hdl == other.hdl) &&  //
           (id == other.id) &&    //
           (ptr == other.ptr) &&  //
           (typ == other.typ);
  }
};
// struct related
struct Hdl_id {
  string        hdl;
  unsigned long id;

  Hdl_id(Hip const& hip) : hdl{hip.hdl}, id{hip.id} {}

  auto operator==(Hdl_id const& other) -> bool {
    return hdl == other.hdl &&  //
           id == other.id;
  }
};
// put-to op related
inline std::ostream& operator<<(std::ostream& os, Hdl_id const& hid) {
  return os << hid.hdl << '.' << hid.id;
}

//------------------------------------------------------------------------------
/** Prints a Hip tag's elements out to the console, with an optional message
 * @param[in] hip the Hip data struct to print
 * @param[in] msg optional description (&tc) string
 */
void prt_tag(Hip const& tag, std::string const& msg = "");

//------------------------------------------------------------------------------
/** Our alias for a standard @c system_clock @c time_point
 */
using Time = time_point<system_clock>;
// alias related:
using Call_clk = system_clock;

//------------------------------------------------------------------------------
/** Provides a mechanism for using @c Hip 's to trace classes' member function
 * calls. Collections of @c stack<> 's of these == *RW system callgraphs*
 * data.
 * @note: @c Hip @b tag, @c string @b func_nm, @c Time @b t
 */
struct Callstamp {
  Hip    tag;
  string func_nm;
  Time   tim;
};

//------------------------------------------------------------------------------
/** Our alias for a stack of @c Callstamp 's
 */
using Stampstack = stack<Callstamp, vector<Callstamp>>;

//------------------------------------------------------------------------------
/** A structure for @c Hip -tagged @c Stampstack 's
 */
struct RW_stamps {
  Hip        rw;
  Stampstack stack;
};

//------------------------------------------------------------------------------
/** Our alias for a collection of @c Hip -tagged @c Stampstack 's
 */
using Callgraph = vector<RW_stamps>;

//------------------------------------------------------------------------------
/** Perform call analyses of the active @c RW object's function interactions
 * @note: for now, this simply dumps the entire @c Callgraph out to the console,
 * with no further analysis provided
 */
void analyze_callgraph();

//------------------------------------------------------------------------------
/** Appends a nanosecond-scaled breakdown of the passed in @c Time @a tim
 * @param[in/out] os the stream to append to
 * @param[in] tim the @c Time to display
 * @return the now-appended @c ostream
 */
inline std::ostream& operator<<(std::ostream& os, Time const& tim) {  //

  auto time_slice = tim.time_since_epoch();

  // UTC: +8:00
  using Days = duration<int, ratio_multiply<hours::period, ratio<8>>::type>;

  Days days = duration_cast<Days>(time_slice);
  time_slice -= days;
  auto hrs = duration_cast<hours>(time_slice);
  time_slice -= hrs;
  //
  auto mins = duration_cast<minutes>(time_slice);
  time_slice -= mins;
  auto secs = duration_cast<seconds>(time_slice);
  time_slice -= secs;
  auto millis = duration_cast<milliseconds>(time_slice);
  time_slice -= millis;
  auto micros = duration_cast<microseconds>(time_slice);
  time_slice -= micros;
  auto nanos = duration_cast<nanoseconds>(time_slice);
  //
  return os << dec << setw(2) << mins.count() << "m:" << setw(2) << secs.count()
            << "s:" << setw(3) << millis.count() << "ms:" << setw(3)
            << micros.count() << "us:" << setw(3) << nanos.count()
            << "ns  past hr";

  // https://en.cppreference.com/w/cpp/chrono/duration
  // https://en.cppreference.com/w/cpp/chrono/duration/duration_cast
  // https://en.cppreference.com/w/cpp/numeric/ratio/ratio
  // https://en.cppreference.com/w/cpp/numeric/ratio/ratio_multiply
}

//------------------------------------------------------------------------------
inline void error(std::string const& s) {
  ostringstream ss{};
  ss << "ERROR:\n" << s << '\n';
  throw runtime_error(ss.str());
}

//------------------------------------------------------------------------------
inline void error(std::string const& err_msg,
                  Hip const&         tag,
                  std::string const& func_nm = "") {
  ostringstream ss{};
  ss << "ERROR:\n  " << err_msg << "  [from " << tag.hdl << '.' << tag.id
     << ", via " << func_nm << "]\n";
  throw runtime_error(ss.str());
}

//------------------------------------------------------------------------------
/** All *slave object* classes' guarded functions should error-out on any
 * @a 'non-Master' attempt
 * @param[in] master
 * @param[in] stranger
 * @param[in] slave
 * @param[in] func_nm
 * @param[in] complaint optional
 * @note: wrapper that throws a @c std::runtime_error
 */
inline void only_master_can(
    Hip const&         master,
    Hip const&         stranger,
    Hip const&         slave,
    std::string const& func_nm,
    std::string const& complaint = "Only Master can ask me to do that") {
  puts("");
  prt_tag(master, "Master  \t");
  prt_tag(stranger, "Stranger\t");
  error(complaint, slave, func_nm);
}
//
/** Wrapper that sends worker complaints on to @c only_master_can() error func
 * @param boss
 * @param stranger
 * @param worker
 * @param func_nm
 * @note: wrapper that throws a @c std::runtime_error
 */
// inline void only_boss_can(Hip const&         boss,
//                           Hip const&         stranger,
//                           Hip const&         worker,
//                           std::string const& func_nm) {
// DESIGN: TODO: use a container of Hip 's (bosses) here. Same for
// relations.
// - only_bosses_can()
//
// only_master_can(boss, stranger, worker, func_nm,
//                 "I can only do that for Boss");
// }
//
// DESIGN: TODO:
// only_relations_can(){}

//------------------------------------------------------------------------------
// THE RW 'SWITCHBOARD:A' - library class' fwd decls/aliases
//
#include "action/action_sb.hpp"
#include "animeyes/animeyes_sb.hpp"
#include "arigato/arigato_sb.hpp"
#include "atashi/atashi_sb.hpp"
#include "bones/bones_sb.hpp"
#include "chibitsu/chibitsu_sb.hpp"
#include "curator/curator_sb.hpp"
#include "dollhouse/dollhouse_sb.hpp"
#include "electro/electro_sb.hpp"
#include "gears/gears_sb.hpp"
#include "graphicom/graphicom_sb.hpp"
#include "hearsay/hearsay_sb.hpp"
#include "meido/meido_sb.hpp"
#include "shell/shell_sb.hpp"

//------------------------------------------------------------------------------
// THE RW 'SWITCHBOARD:B' - library header includes
//
#include "rw/rw_incls_all.hpp"  // spcfcly. #include'd before other rw classes
//
#include "action/action_incls_all.hpp"
#include "animeyes/animeyes_incls_all.hpp"
#include "arigato/arigato_incls_all.hpp"
#include "atashi/atashi_incls_all.hpp"
#include "bones/bones_incls_all.hpp"
#include "chibitsu/chibitsu_incls_all.hpp"
#include "curator/curator_incls_all.hpp"
#include "dollhouse/dollhouse_incls_all.hpp"
#include "electro/electro_incls_all.hpp"
#include "gears/gears_incls_all.hpp"
#include "graphicom/graphicom_incls_all.hpp"
#include "hearsay/hearsay_incls_all.hpp"
#include "meido/meido_incls_all.hpp"
#include "shell/shell_incls_all.hpp"

//    --=^=--

// "So is it true Sempai? Are robowaifus really better than any human woman?"

// rw_foundations_lib.v211211
// Copyright (2021)
// License MIT(Expat) https://opensource.org/licenses/MIT
